import platform
import subprocess
from ..constants import *
from ..layers import *
from ..utils import *
from ..materials import *
import json
import gdsfactory as gf
from copy import deepcopy
# from time import time
import datetime
from math import cos, pi, sin
import os
import numpy as np

from sortedcontainers import SortedDict, SortedSet
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def setup(path, study, nres, center_wavelength,
          component=None,   zlims=None, zmargin=None,
          port_margin="auto",
          runs=[],  sources=[],
          layer_stack=SOI, materials=dict(),
          default_material='SiO2',
          exclude_layers=[], Courant=None,
          gpu=None, dtype=np.float32,
          plot=False, saveat=2.5,
          magic="", wd=os.path.join(os.getcwd(), "runs"), name=None,
          source_port_margin=.1,
          Ttrans=None,
          approx_2D_mode=False, show_field='Hz',):
    materials = {**MATERIALS, **materials}
    prob = {
        'nres': nres,
        'center_wavelength': center_wavelength,
        'name': name,
        'path': path,
        'show_field': show_field,
    }
    if approx_2D_mode:
        N = 2
        prob["approx_2D_mode"] = approx_2D_mode
    else:
        N = 3
        prob["approx_2D_mode"] = None
    prob["class"] = "pic"
    prob["Ttrans"] = Ttrans
    prob["dtype"] = str(dtype)
    prob["timestamp"] = datetime.datetime.now().isoformat(
        timespec="seconds").replace(":", "-")
    prob["magic"] = magic
    prob["saveat"] = saveat

    gpu_backend = gpu
    # if gpu_backend:s
    prob["gpu_backend"] = gpu_backend

    if component is None:
        0
    else:
        c = component
        ports = {
            p.name: {
                "center": (np.array(p.center)/1e0).tolist(),
                "normal": [cos(p.orientation/180*pi), sin(p.orientation/180*pi)],
                "tangent": [-sin(p.orientation/180*pi), cos(p.orientation/180*pi)],
                'width': p.width/1e0,
            }
            for p in c.get_ports_list(prefix="o")
        }
        for (k, v) in ports.items():
            z, n = [0, 0, 1], [*v['normal'], 0]
            t = np.cross(z, n).tolist()
            v['frame'] = [t, z, n]
            # v["dimensions"] = [v['width']]

        prob["ports"] = ports
        mode_solutions = []

        d = layer_stack.layers['core']
        hcore = d.thickness
        zcore = d.zmin

        if zmargin is None:
            zmargin = 4*hcore
        if type(zmargin) in [int, float]:
            zmargin = [zmargin, zmargin]

        h = hcore+zmargin[0]+zmargin[1]
        zmin = zcore-zmargin[0]
        zmax = zmin+h
    #
        port_width = max([p.width/1e0 for p in c.ports])
        # xmargin = ymargin = 2*port_width

        modexmargin = port_width
        modezmargin = [.8*zmargin[0], .8*zmargin[1]]
        wmode = port_width+2*modexmargin
        hmode = hcore+modezmargin[0]+modezmargin[1]
        zmode = zcore-modezmargin[0]
        zcenter = zmode+hmode/2

        prob["hcore"] = hcore
        prob["zcenter"] = zcenter
        prob["zmin"] = zmin
        prob["zmax"] = zmax
        prob["zcore"] = zcore
        # prob["L"] = [l, w, h]

        layers = set(c.layers)-set(exclude_layers)

        MODES = os.path.join(path, "modes")
        os.makedirs(MODES, exist_ok=True)
        GEOMETRY = os.path.join(path, 'geometry')
        os.makedirs(GEOMETRY, exist_ok=True)

        layer_stack_info = material_voxelate(
            c,  zmin, zmax, layers, layer_stack, GEOMETRY)
        dir = os.path.dirname(os.path.realpath(__file__))

        for f in ["solvemodes.py"]:
            fn = os.path.join(dir, f)
            if platform.system() == "Windows":
                os.system(f"copy /Y {fn} {MODES}")
            else:
                subprocess.run(["cp", fn, MODES])
        prob["layer_stack"] = layer_stack_info
        prob["study"] = study
        prob["materials"] = materials

        prob["N"] = N

        prob["hmode"] = hmode
        prob["wmode"] = wmode
        prob["zmode"] = zmode
        wmode = port_width+2*modexmargin
        # _c = add_bbox(c, layer=BBOX, nonport_margin=margin)
        for run in runs:
            for k, v in list(run["sources"].items())+list(run["monitors"].items()):
                p = ports[k]
                # v['dimensions'] = [wmode, hmode]
                v['start'] = [-wmode/2, zmode-zcenter]
                v['stop'] = [wmode/2, zmode+hmode-zcenter]
                v['frame'] = p['frame']
                # for center_wavelength in v["wavelength_mode_numbers"]:
                #     wavelengths.append(center_wavelength)
            for k, v in run['sources'].items():
                p = ports[k]
                ct = np.array(p['center'])
                n = np.array(p['normal'])
                v["center"] = (ct+n*source_port_margin).tolist()
                v['center'] += [zcenter]
            for k, v in run['monitors'].items():
                p = ports[k]
                v['center'] = copy.deepcopy(p['center'])
                v['center'] += [zcenter]

        prob["mode_solutions"] = mode_solutions
        prob["runs"] = runs
        bbox = c.bbox_np().tolist()
        bbox[0].append(zmin)
        bbox[1].append(zmax)

        prob['bbox'] = bbox
    # prob['epdefault'] = materials[layer_stack['default']['material']]['epsilon']
    prob['epdefault'] = materials[default_material]['epsilon']

    prob["Courant"] = Courant
    if not os.path.exists(path):
        os.makedirs(path)
    prob["mode_solutions"] = mode_solutions
    # prob["path_length"] = 2*(l+r+lwg)
    return prob


def port_name(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        return s
    return f"o{s}"


def port_number(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        s = s[1:]
    return int(s)


def mode_number(port):
    l = str(port).split("@")
    return 0 if len(l) == 1 else int(l[1])


def unpack_sparam_key(k):
    o, i = k.split(",")
    po, pi = port_name(o), port_name(i)
    mo, mi = mode_number(o), mode_number(i)
    return po, mo, pi, mi


def long_sparam_key(k):
    po, mo, pi, mi = unpack_sparam_key(k)
    return f"{po}@{mo},{pi}@{mi}"



from IPython.display import Image
from PIL import Image as PILImage
import os
# import pymeshfix
import json
from statistics import median
from .constants import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from .constants import *
from .materials import *
try:
    from IPython.display import display
except ImportError:
    pass
# import pyvista as pv

# from .picToGDS import main

from math import ceil, cos, pi, sin, tan
import matplotlib.pyplot as plt
import numpy as np
from gdsfactory.generic_tech import LAYER_STACK, LAYER
import copy
import shutil
import trimesh
# from gdsfactory import LAYER_VIEWS

tol = .001


def get(c, i):
    try:
        return c[i]
    except:
        try:
            return getattr(c, i)
        except:
            return None


def arange(a, b, d):
    ret = np.linspace(a, b, round((b-a)/(d))+1).tolist()
    return ret


def trim(x, dx):
    return round(x/dx)*dx


def extend(endpoints, wm):
    v = endpoints[1]-endpoints[0]
    v = v/np.linalg.norm(v)
    return [
        (endpoints[0]-wm*v).tolist(), (endpoints[1]+wm*v).tolist()]


def portsides(c):
    ports = c.ports
    bbox = c.bbox_np()
    res = [[], [], [], []]
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    for p in ports:
        x, y = np.array(p.center)/1e0

        if abs(x - xmin0) < tol:
            res[2].append(p.name)
        if abs(x - xmax0) < tol:
            res[0].append(p.name)
        if abs(y - ymin0) < tol:
            res[3].append(p.name)
        if abs(y - ymax0) < tol:
            res[1].append(p.name)
    return res


def add_bbox(c, layer, nonport_margin=0):  # , dx=None):
    bbox = c.bbox_np()
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    l = xmax0-xmin0
    w = ymax0-ymin0

    # l = dx*np.ceil((xmax0-xmin0)/dx)
    # w = dx*np.ceil((ymax0-ymin0)/dx)

    # if dx is not None:
    #     if nonport_margin is None:
    #         nonport_margin = dx
    # if nonport_margin is None:
    #     nonport_margin = 0
    margin = nonport_margin
    xmin, ymin, xmax, ymax = xmin0-margin, ymin0 - \
        margin, xmin0+l+margin, ymin0+w+margin

    for p in c.ports:
        # p = c.ports[k]
        x, y = np.array(p.center)/1e0
        if abs(x - xmin0) < tol:
            xmin = x
        if abs(x - xmax0) < tol:
            xmax = x
        if abs(y - ymin0) < tol:
            ymin = y
        if abs(y - ymax0) < tol:
            ymax = y
    p = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
    _c = gf.Component()
    _c << c

    if type(layer[0]) is int:
        layer = [layer]
    for layer in layer:
        layer = tuple(layer)
        _c.add_polygon(p, layer=layer)
    for port in c.ports:
        _c.add_port(name=port.name, port=port)
    return _c

# def pic2gds(fileName, sizeOfTheCell, layerNum=1, isDither=False, scale=1):
    main(fileName, sizeOfTheCell, layerNum, isDither, scale)
    return "image.bmp", "image.gds"


def finish(c, name):
    c.add_label(name, position=c.bbox_np()[1])


def normal_from_orientation(orientation):
    return [cos(orientation/180*pi), sin(orientation/180*pi)]


def material_voxelate(c,  zmin, zmax, layers, layer_stack, path):
    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path, exist_ok=True)
    stacks = sum([[[v.mesh_order, v.material, tuple(layer), k]
                 for k, v in get_layers(layer_stack, layer, withkey=True)] for layer in layers], [])
    c.flatten()
    stacks = sorted(stacks, key=lambda x: -x[0])
    layer_stack_info = dict()
    # raise NotImplementedError("This is a stub")
    lb, ub = c.bbox_np()
    # bbox = [[**lb, zmin], [**ub, zmax]]
    bbox = [[lb[0], lb[1], zmin], [ub[0], ub[1], zmax]]
    layers = [x[2] for x in stacks]

    layer_views = copy.deepcopy(LAYER_VIEWS)

    for i, stack in enumerate(stacks):
        order = stack[0]
        m = stack[1]
        l1, l2 = layer = stack[2]
        k = stack[3]

        layer_views.layer_views[k] = LayerView(
            layer=layer, color='blue', visible=True)
        _layer_stack = copy.deepcopy(layer_stack)
        get(_layer_stack, 'layers').clear()

        d = copy.deepcopy(get(layer_stack, 'layers')[k])
        if d.zmin <= zmax and d.bounds[1] >= zmin:
            get(_layer_stack, 'layers')[k] = d
            d.zmin = max(zmin, d.zmin)
            d.thickness = min(zmax-d.zmin, d.thickness)
            # _d.bounds = (_d.zmin, _d.zmin+_d.thickness)
            # origin = (c.extract([layer]).bbox_np()-c.bbox_np())[0].tolist()

            mesh = c.to_3d(
                layer_stack=_layer_stack,
                layer_views=layer_views,
                exclude_layers=set(layers)-{layer})
            OBJ = os.path.join(path, f'{order}_{m}_{k}.obj')
            trimesh.exchange.export.export_mesh(mesh, OBJ, 'obj')
            # pymeshfix.clean_from_file(OBJ, OBJ)
            # repair_obj_open3d(OBJ, OBJ)

            # STL = os.path.abspath(os.path.join(path, f"{k}.stl"))
            # gf.export.to_stl(c, STL, layer_stack=_layer_stack)

            # STL = os.path.join(path, f'{k}_{l1}_{l2}.stl')
            # pymeshfix.clean_from_file(STL, STL)
            # STL1 = os.path.join(path, f'{order}_{m}_{k}.stl')
            # os.replace(STL, STL1)

            # mesh = pv.read(STL)
            # im = stl_to_array(mesh, dl*unit, bbox)
            # np.save(os.path.join(path, f'{k}.npy'), im)

            layer_stack_info[k] = {
                "layer": (l1, l2),
                "zmin": d.zmin,
                "thickness": d.thickness,
                "material": matname(m),
                "mesh_order": stack[0],
                # "origin": origin,
            }
    return layer_stack_info


def get_layers(layer_stack, layer, withkey=False):
    r = []
    d = get(layer_stack, 'layers').items()

    for k, x in d:
        l = get(x, 'layer')
        if l is not None:
            t = get(l, 'layer')
            if t is not None and tuple(t) == tuple(layer):
                if withkey:
                    x = k, x
                r.append(x)
    if r:
        return r

    for k, x in d:
        l = get(x, 'derived_layer')
        if l is not None:
            t = get(l, 'layer')
            if t is not None and tuple(t) == tuple(layer):
                if withkey:
                    x = k, x
                r.append(x)
    return r


def wavelength_range(center, bandwidth, length=3):
    f1 = 1/(center+bandwidth/2)
    f2 = 1/(center-bandwidth/2)
    hw = (f2-f1)/2
    f1 = 1/center-hw
    f2 = 1/center+hw
    return sorted([1/x for x in np.linspace(f1, f2, length).tolist()])


def wrap(wavelengths):
    if isinstance(wavelengths, float) or isinstance(wavelengths, int):
        wavelengths = [[wavelengths]]
    elif isinstance(wavelengths[0], float) or isinstance(wavelengths[0], int):
        wavelengths = [wavelengths]
    return wavelengths


def save_problem(prob, path):
    path = os.path.abspath(path)
    bson_data = json.dumps(prob)
    # prob["component"] = c0

    path = prob["path"]
    if not os.path.exists(path):
        os.makedirs(path)
        #   compiling julia code...
        #   """)
    prob_path = os.path.join(path, "problem.json")
    with open(prob_path, "w") as f:
        # Write the BSON data to the file
        f.write(bson_data)
    print("using simulation folder", path)


def load_prob(path):
    path = os.path.abspath(path)
    print(f"loading problem from {path}")
    return json.loads(open(os.path.join(path, "problem.json"), "rb").read())


def create_gif(image_path, output_path, duration):
    """
    Creates a GIF from a list of image paths.

    Args:
        image_paths: A list of file paths to the images.
        output_path: The output path for the generated GIF.
        duration: The duration of each frame in milliseconds (default: 200).
    """
    image_paths = os.listdir(image_path)
    image_paths = sorted(image_paths, key=lambda x: float(x[0:-4]))
    frames = [PILImage.open(os.path.join(image_path, f))
              for f in image_paths]

    # Ensure all frames have the same palette if necessary
    for i in range(len(frames)):
        if frames[i].mode != "RGB":
            frames[i] = frames[i].convert("RGB")

    frames[0].save(
        output_path,
        save_all=True,
        append_images=frames[1:],
        duration=duration,
        loop=0  # 0 means infinite loop
    )


def make_movie(path):
    framerate = 10/load_prob(path)["saveat"]
    duration = 1000/framerate
    FRAMES = os.path.join(path, "frames")
    MOVIE = os.path.join(path, "sim.mp4")
    # shutil.rmtree(MOVIE, ignore_errors=True)
    GIF = os.path.join(path, "sim.gif")
    create_gif(FRAMES, GIF, duration)
    return Image(open(GIF, 'rb').read())
    # if saveat is None:
    #     saveat = load_prob(path)["saveat"]

    # g = np.load(os.path.join(path, "temp", 'g.npy')).T
    # gmax = np.max(np.abs(g))

    # dir = os.path.join(path, "temp", 'fields')
    # umax = 0
    # fns = sorted(os.listdir(dir), key=lambda x: float(x[0:-4]))
    # for fn in fns:
    #     a = np.load(os.path.join(dir, fn))
    #     v = np.max(np.abs(a))
    #     if umax < v:
    #         umax = v

    # for fn in fns:
    #     name = fn[0:-4]
    #     a = np.load(os.path.join(FIELDS, fn)).T
    #     fig, axs = plt.subplots(1, 2)
    #     axs[1].imshow(a, cmap='seismic', origin='lower',
    #                   vmin=-umax, vmax=umax)
    #     axs[0].imshow(-g, cmap='gray',
    #                   origin='lower', vmin=-gmax, vmax=0)
    #     plt.savefig(os.path.join(FRAMES, f"{name}.png"))
    #     plt.close(fig)

    # fns = sorted(os.listdir(FRAMES), key=lambda x: float(x[0:-4]))
    # frame = height = width = layers = video = None
    # for i, fn in enumerate(fns):
    #     frame = cv2.imread(os.path.join(FRAMES, fn))
    #     if i == 0:
    #         height, width, layers = frame.shape
    #         video = cv2.VideoWriter(
    #             MOVIE, 0x7634706d, saveat, (width, height))

    #     video.write(frame)

    # cv2.destroyAllWindows()
    # video.release()




from scipy.special import comb
from ..constants import *
from ..utils import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from networkx import center
from numpy import cumsum
from gdsfactory.generic_tech import LAYER_STACK, LAYER

# import .utils as utils
# from layers import *


def port_bbox(p):
    c = np.array(p.center)
    v = p.width/2*np.array([np.cos(np.radians(p.orientation+90)),
                           np.sin(np.radians(p.orientation+90))])
    return [(c-v).tolist(), (c+v).tolist()]


def bezier_curve(points, num_points=100):
    n = len(points) - 1
    t = np.linspace(0, 1, num_points)
    curve_points = np.zeros((num_points, 2))

    for i, t_val in enumerate(t):
        for j, point in enumerate(points):
            curve_points[i] += comb(n, j) * (t_val ** j) * \
                ((1 - t_val) ** (n - j)) * point

    return curve_points


def mimo(l, w,
         west=0, east=0, south=0, north=0,
         lwg_in=None, lwg_out=None, taper=0.04,
         wwg=.5,  wwg_west=None, wwg_east=None, wwg_south=None, wwg_north=None,
         ports_in=[],
         layer=LAYER.WG,           canvas_layer=CANVAS_LAYER,
         **kwargs):
    design = gf.Component()
    c = gf.Component(**kwargs)
    if lwg_in is None:
        lwg_in = wwg
    if lwg_out is None:
        lwg_out = 2*wwg

    p = [(0, 0), (l, 0), (l, w), (0, w)]
    design.add_polygon(p,                       layer=canvas_layer)
    # c.add_polygon(p,                       layer=layer)

    port_pos_sides = [west, north,  east, south]
    for i, v, d in zip(range(4), port_pos_sides, [w, w, l, l]):
        if type(v) is int:
            port_pos_sides[i] = [(.5+j)*d/v for j in range(v)]

    wwg_sides = [wwg_west, wwg_east, wwg_south, wwg_north]
    for i, v in enumerate(wwg_sides):
        n = len(port_pos_sides[i])
        if v is None:
            v = wwg
        if type(v) is float or type(v) is int:
            wwg_sides[i] = [v]*n

    nports = 0
    for (i, x, y, ds, wwgs, orientation) in zip(
        range(4),
        [0,  0, l, l],
        [0, w, w, 0],
        port_pos_sides,
        wwg_sides,
        [180, 90, 0, -90]
    ):
        for wwg, d in zip(wwgs, ds):
            center = [x+cos(np.radians(orientation-90))*d, y +
                      sin(np.radians(orientation-90))*d]
            nports += 1
            if nports in ports_in:
                wwg2 = wwg+taper*lwg_in
                lwg = lwg_in
            else:
                wwg2 = wwg+taper*lwg_out
                lwg = lwg_out

            name = "o"+str(nports)
            design.add_port(name, center=center, width=wwg2,
                            orientation=orientation, layer=layer)
            wg = c << gf.components.taper(
                length=lwg, width1=wwg, width2=wwg2, layer=layer)
            wg.connect("o2", design.ports[name])
            c.add_port(name, port=wg.ports["o1"])

    design = c << design
    p = []
    for i in range(nports):
        # for i in [0]:
        pi = design.ports[f'o{i+1}']
        pj = design.ports[f'o{((i+1) % nports)+1}']
        a, _ = np.array(port_bbox(pi))
        _, b = np.array(port_bbox(pj))
        n1 = - np.array([cos(np.radians(pi.orientation)),
                         sin(np.radians(pi.orientation))])
        n2 = -np.array([cos(np.radians(pj.orientation)),
                       sin(np.radians(pj.orientation))])
        v = b-a
        d = np.linalg.norm(v)
        n = v/d

        # l=[a, a+.3*d * (.5*n+.5*n1),  b+.3*d*(-.5*n+.5*n2), b]
        l = [a, a+.5*d * n1,  b+.5*d*n2, b]
        p.extend(bezier_curve(l))

        # p.extend([a+s*d * (2*s*n+(1-2*s)*n1) for s in np.linspace(0, .5, 50)])
        # p.extend(reversed([b+s*d * (-2*s*n+(1-2*s)*n2)
        #          for s in np.linspace(0, .5, 50)]))
        # c << gf.components.bends.bezier(
        #     [x.tolist() for x in l],
        #     # start_angle=pi.orientation-180,
        #     # end_angle=pj.orientation,
        #     allow_min_radius_violation=True,
        #     width=pi.width,)  # layer=layer)
    c.add_polygon(p, layer=layer)
    return c



from .setup import *
from ..constants import *
from ..layers import *
from ..utils import *
import gdsfactory as gf
from copy import deepcopy
import matplotlib.pyplot as plt
import numpy as np

from gdsfactory.cross_section import Section
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def make_prob(path,  nres,
              component=None,
              wavelengths=None,
              entries=None, keys=["2,1"],
              layer_stack=SOI,
              study="sparams",
              sources=None, monitors=None,
              frequencies=None,   reference_wavelength=None,
              source_port_margin=.1,
              T=None,
              **kwargs):
    if frequencies is not None:
        frequencies = wrap(frequencies)
        assert reference_wavelength is not None
        assert wavelengths is None
        wavelengths = [[reference_wavelength /
                       f for f in v] for v in frequencies]
    else:
        wavelengths = wrap(wavelengths)

    wavelengths = [sorted(w) for w in wavelengths]
    wavelengths = sorted(wavelengths, key=lambda x: x[0])
    center_wavelength = median(map(median, wavelengths))

    if not entries:
        entries = []

        for w in wavelengths:
            for k in keys:
                entries.append([w, *unpack_sparam_key(k)])

    l = []
    for w, po, mo, pi, mi in entries:
        k = [w, pi, mi, pi, mi]
        if k not in entries:
            l.append(k)
    entries.extend(l)

    imow = {}
    for w, po, mo, pi, mi in entries:
        if pi not in imow:
            imow[pi] = {}
        if mi not in imow[pi]:
            imow[pi][mi] = {}

        if po not in imow[pi][mi]:
            imow[pi][mi][po] = mo
        else:
            imow[pi][mi][po] = max(imow[pi][mi][po], mo)

    runs = []
    for _w in [1]:
        for i in imow:
            for mi in imow[i]:
                d = {
                    "sources": {
                        i: {
                            "wavelength_mode_numbers": [[w, [mi]] for w in wavelengths],
                        }},
                    "monitors": {
                        o: {
                            "wavelength_mode_numbers": [[w, list(range(imow[i][mi][o]+1))] for w in wavelengths],
                        } for o in imow[i][mi]}}
                d["sources"] = SortedDict(d["sources"])
                d["monitors"] = SortedDict(d["monitors"])
                if component is None:
                    0
                else:
                    runs.append(d)

    prob = setup(path, component=component, study=study,  nres=nres, center_wavelength=center_wavelength,
                 runs=runs,
                 source_port_margin=source_port_margin,
                 #  sources=sources, monitors=monitors,
                 layer_stack=layer_stack, **kwargs)
    prob["wavelengths"] = wavelengths
    prob["T"] = T
    save_problem(prob, path)
    return prob

    # l = [k for k in imow if port_number(k) == pi]
    # if not l:
    #     imow[f"o{pi}@{mi}"] = []
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[i] = imow[k]
    #         del imow[k]

    # l = [k for k in imow[i] if port_number(k) == po]
    # if not l:
    #     imow[f"o{pi}@{mi}"]
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[f"o{pi}@{mn}"] = imow[k]
    #         del imow[k]

    # if po not in imow[pi]:
    #     imow[pi]["o"][po] = mo
    # else:
    #     imow[pi]["o"][po] = max(imow[pi]["o"][po], mo)
